"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Schema = void 0;
const common_1 = require("./common");
class Schema {
    constructor(schema) {
        this.types = schema && schema.types || [];
        this.nodes = schema && schema.nodes || [];
    }
    add(type, value, index = this[type].length) {
        const patch = { op: "add", path: `/${type}/${index}`, value: [...value] };
        if (type === "nodes" && typeof value[3] === "string") {
            // value[3] is map key => need to add key to parent map
            const parentIndex = this.nodes.findIndex((n) => n[0] === value[2]);
            const keyIndex = this.nodes[parentIndex].findIndex((n) => n === value[3]);
            // check if key already exists
            if (keyIndex < 4) {
                // add key to map node
                this.addMapNodeKey(parentIndex, value[3]);
                // put key index in value[3]
                value[3] = this.nodes[parentIndex].length - 5;
            }
            else {
                // put key index in value[3]
                value[3] = keyIndex - 4;
            }
        }
        // add value to schema
        this[type].splice(index, 0, value);
        return patch;
    }
    replace(type, index, value) {
        this[type][index] = value;
        return { op: "replace", path: `/${type}/${index}`, value, };
    }
    remove(type, index) {
        this[type].splice(index, 1);
        return { op: "remove", path: `/nodes/${index}` };
    }
    addType(name, props) {
        common_1.check(this.types.find((n) => n[0] === name), `Cannot add type to schema - type ${name} already exists!`);
        return this.add("types", [name, ...props]);
    }
    addArrayNode(nodeId, parentId, name) {
        return this.addNode(nodeId, common_1.NODE_ARRAY_TYPE, parentId, name);
    }
    addMapNode(nodeId, parentId, name, keys = []) {
        return this.addNode(nodeId, common_1.NODE_MAP_TYPE, parentId, name, keys);
    }
    addObjectNode(nodeId, type, parentId, name) {
        const typeIndex = this.types.findIndex((n) => n[0] === type);
        common_1.check(typeIndex < 0, `Cannot add node to schema - node type ${type} is not found!`);
        return this.addNode(nodeId, typeIndex, parentId, name);
    }
    addNode(nodeId, type, parentId, name, keys = []) {
        common_1.check(this.getNode(nodeId), `Cannot add node to schema - node with id ${nodeId} already exists!`);
        common_1.check(parentId === -1 && this.nodes.length, `Cannot add second root node to schema - only root node can have parentId -1!`);
        const parent = this.getNode(parentId);
        common_1.check(!parent && parentId >= 0, `Cannot add node to schema - parent with id ${parentId} not found!`);
        const index = typeof name === "string" && (!parent || parent.type !== common_1.NODE_MAP_TYPE)
            ? this.getNodeIndex(parentId, name) : name;
        common_1.check(index === -1 && this.nodes.length, `Cannot add node to schema - key/index "${name}" not exist!`);
        return this.add("nodes", [nodeId, type, parentId, index, ...keys]);
    }
    addMapNodeKey(index, key) {
        const node = this.nodes[index];
        common_1.check(!node, `Cannot add key to schema - node with index ${index} not found`);
        common_1.check(node.slice(4).indexOf(key) >= 0, `Cannot add key to schema - key ${key} already exists!`);
        node.push(key);
        return { op: "add", path: `/nodes/${index}/${node.length - 1}`, value: key };
    }
    deleteNode(nodeId, deleteChildren = false) {
        const index = this.nodes.findIndex((n) => n[0] === nodeId);
        common_1.check(index < 0, `Cannot delete node to schema - node with id ${nodeId} not exists!`);
        if (deleteChildren) {
            this.nodes.filter((n) => n[2] === nodeId).forEach((n) => this.deleteNode(n[0]));
        }
        return this.remove("nodes", index);
    }
    getTypeProps(index) {
        const type = this.types[index];
        return type ? type.slice(1) : [];
    }
    getNodeName(sn) {
        const node = Array.isArray(sn) ? common_1.schemaNode(sn) : sn;
        const parent = this.getNode(node.parent);
        if (!parent) {
            return "";
        }
        else if (parent.type === common_1.NODE_ARRAY_TYPE) {
            return node.index;
        }
        else if (parent.type === common_1.NODE_MAP_TYPE) {
            return parent.items[node.index];
        }
        else {
            const { props } = common_1.schemaType(this.types[parent.type]);
            return props[node.index];
        }
    }
    getNodeIndex(parentId, name) {
        const parent = this.getNode(parentId);
        if (!parent) {
            return -1;
        }
        else if (parent.type === common_1.NODE_MAP_TYPE) {
            return parent.items.indexOf(name);
        }
        else if (parent.type === common_1.NODE_ARRAY_TYPE) {
            return +name;
        }
        else {
            const props = this.getTypeProps(parent.type);
            return props.indexOf(name);
        }
    }
    getNode(nodeId) {
        const node = this.nodes.find((n) => n[0] === nodeId);
        return node && common_1.schemaNode(node);
    }
    getNodePath(sn) {
        let node = Array.isArray(sn) ? common_1.schemaNode(sn) : sn;
        const pathArr = [];
        while (!!node) {
            pathArr.push(this.getNodeName(node));
            node = this.getNode(node.parent);
        }
        return pathArr.reverse().join("/");
    }
    getChildNode(sn, name) {
        if (!sn) {
            return this.findNode({ parent: -1 });
        }
        const node = Array.isArray(sn) ? common_1.schemaNode(sn) : sn;
        if (node.type === common_1.NODE_ARRAY_TYPE) {
            return this.findNode({ parent: node.id, index: +name });
        }
        else if (node.type === common_1.NODE_MAP_TYPE) {
            return this.findNode({ parent: node.id, index: node.items.indexOf(name) });
        }
        else {
            const { props } = common_1.schemaType(this.types[node.type]);
            return this.findNode({ parent: node.id, index: props.indexOf(name) });
        }
    }
    findNode(query) {
        const sn = this.nodes.find((n) => {
            const node = common_1.schemaNode(n);
            return !Object.keys(query).find((key) => node[key] !== query[key]);
        });
        return sn && common_1.schemaNode(sn);
    }
}
exports.Schema = Schema;
